import inspect
import functools
from typing import OrderedDict, Tuple


def _cast(
        obj: object,
        cast_to: type,
        cast_rules: OrderedDict[Tuple[type, type], callable]):
    for (input_type, output_type), func in cast_rules.items():
        if isinstance(obj, input_type):
            return func(obj)
    else:
        return cast_to(obj)


def _autocast(
        cast_rules: OrderedDict[Tuple[type, type], callable] = {},
        **cast_to: type) -> callable:

    def wrapper(func: callable) -> callable:
        argspec = inspect.getfullargspec(func)
        arg_names = argspec.args
        for key, value in cast_to.items():
            if key not in arg_names:
                raise ValueError(
                    f"arg '{key}' not found in {func.__name__}().")
            if not isinstance(value, type):
                raise TypeError(
                    f"Cast destination of arg '{key}' in {func.__name__}() "
                    "must be an instance of type.")

        @functools.wraps(func)
        def _func_with_typecast(*args, **kwargs):
            args_casted = []
            for name, arg in zip(arg_names, args):
                if name in cast_to:
                    args_casted.append(_cast(arg, cast_to[name], cast_rules))
                else:
                    args_casted.append(arg)
            kwargs_casted = {}
            for name, arg in kwargs.items():
                if name in cast_to:
                    kwargs_casted[name] = _cast(arg, cast_to[name], cast_rules)
                else:
                    kwargs_casted[name] = arg
            return func(*args_casted, **kwargs_casted)

        _func_with_typecast.__signature__ = inspect.signature(func)
        return _func_with_typecast

    return wrapper


def autocast(**cast_to: type) -> callable:
    """Decorator to automatically cast function arguments.

    Parameters
    ----------
    cast_to : type
        Specifies casting destination of arguments.

    Returns
    -------
    callable
        Function with type casting.

    Example
    -------
    >>> from pyautocast import autocast
    >>> @autocast(x=str)
    ... def func(x):
    ...     assert(isinstance(x, str))
    ...     return "arg 'x' in func() is " + x
    ...
    >>> func(2)
    "arg 'x' in func() is 2"
    >>> func([1, 2, 3])
    "arg 'x' in func() is [1, 2, 3]"
    """
    return _autocast({}, **cast_to)



from setuptools import setup, find_packages


__version__ = '0.1.1'


develop_requires = [
    'autopep8',
    'flake8',
    'pep8-naming',
]


try:
    with open('README.md') as f:
        readme = f.read()
except IOError:
    readme = ''


setup(
    name='pyautocast',
    version=__version__,
    author='ctgk',
    author_email='r1135nj54w@gmail.com',
    maintainer='ctgk',
    maintainer_email='r1135nj54w@gmail.com',
    url='https://github.com/ctgk/pyautocast',
    description='Automatic function arguments casting',
    long_description=readme,
    long_description_content_type='text/markdown',
    license="MIT",
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: MIT License',
    ],

    packages=find_packages(exclude=('test*',)),
    python_requires=">=3.7",
    extras_require={"develop": develop_requires},

    test_suite="tests",
)



from pyautocast._autocast import _autocast, autocast


class CustomCast(object):
    """Class to perform custom type casting.

    Example
    -------
    >>> from pyautocast import CustomCast
    >>> mycast = CustomCast()
    >>> mycast.add_cast_rule(int, tuple, lambda x: (x, x))
    >>> @mycast.autocast(x=tuple)
    ... def func(x):
    ...     print(x)
    >>> func(2)
    (2, 2)
    >>> func(-4.5)
    Traceback (most recent call last):
    ...
    TypeError: 'float' object is not iterable
    """

    def __init__(self):
        super().__init__()
        self._cast_rules = {}

    def add_cast_rule(
            self,
            input_type: type,
            output_type: type,
            cast_func: callable):
        """Register a type casting rule.

        Parameters
        ----------
        input_type : type
            Type of input argument.
        output_type : type
            Target type to cast to when an instance of `input_type` is given.
        cast_func : callable
            Function to convert an instance of `input_type` to `output_type`.

        Raises
        ------
        TypeError
            Raises if arg `input_type` or `output_type`
            is not an instance of type.
        """
        if not isinstance(input_type, type):
            raise TypeError(
                f"arg `input_type` in {self.add_cast_rule.__name__}() "
                "must be an instance of type.")
        if not isinstance(output_type, type):
            raise TypeError(
                f"arg `output_type` in {self.add_cast_rule.__name__}() "
                "must be an instance of type.")
        if not callable(cast_func):
            raise TypeError(
                f"arg `cast_func` in {self.add_cast_rule.__name__}() "
                "must be callable.")
        self._cast_rules[(input_type, output_type)] = cast_func

    def autocast(self, **cast_to: type):
        return _autocast(self._cast_rules, **cast_to)


__all__ = [
    'autocast',
    'CustomCast',
]
